;+
; NAME:
;
;  epoxi_crrej.pro
;
; PURPOSE:
;
;  Exercise a quick cosmic ray rejection for EPOXI images
;
; CATEGORY:
;
;  DI/EPOXI image processing
;
; CALLING SEQUENCE:
;
;  output = epoxi_crrej(input[, mask][, header=header][, box=box_size]
;                            [, keep_overscan=[0|1]][, /verbose])
;
; INPUTS:
;
;  input    - Input image
;
; OPTIONAL INPUTS:
;
;
;
; KEYWORD PARAMETERS:
;
;  header   - Image header.  It will return the modified header with
;             keyword "CRREJ_JY" added.
;  box      - Size of median box used to replace cosmic rays.  Default
;             is 7.
;  keep_overscan  - A non-zero value will keep the overscan area
;  verbose  - Verbose mode for imgclean
;
; OUTPUTS:
;
;  Program returns the cosmic ray cleaned image.
;
; OPTIONAL OUTPUTS:
;
;  mask     - a byte mask: 0 - pixels not changed; 1 - cosmic rays; 2
;             - overscan area
;
; COMMON BLOCKS:
;
;
;
; SIDE EFFECTS:
;
;
;
; RESTRICTIONS:
;
;
;
; PROCEDURE:
;
;
;
; EXAMPLE:
;
;
;
; MODIFICATION HISTORY:
;
;  10/12/2010, created by JYL @UMd
;    Based on imgcleanf.pro provided with the DI calibration pipeline.
;-

function epoxi_crrej, input, mask, header=header, box=box, keep_overscan=keep, verbose=verbose

  if n_params() lt 1 then begin
    message, /con,'Usage: output = epoxi_crrej(input[, mask][, header=header][, box=box_size][, keep_overscan=[0|1]][, /verbose])'
    return, -1
  endif

  if size(verbose,/type) eq 0 then verbose=0
  if size(box,/type) eq 0 then box=7
  if size(keep,/type) eq 0 then keep=1
  if keep then if size(header,/type) eq 0 then begin
    message, /con,'Warning: NO header available, overscan area may be lost!'
    keep = 0
  endif

  out = input   ; output image
  mask = bytarr(size(input,/dimensions))    ; output mask

  ; detect cosmic rays with imgclean.  used here is the version with
  ; "interactive" keyword added so that the print out to screen is
  ; supressed
  imgclean,out, interactive=verbose

  ; median filtered image
  med=median(input,box)

  ; interpolate over cosmic rays
  ww = where(out ne input,cnt)
  if cnt gt 0 then begin
    out[ww] = med[ww]
    mask[ww] = 1
    ; if the new brightness is greater than old brightness, then
    ; use the old brightness because CRs are brighter than background
    tooBright = where(med[ww] GT input[ww],c2)
    if c2 gt 0 then begin
      out[ww[tooBright]] = input[ww[tooBright]]
      mask[ww[tooBright]] = 0
    endif
  endif

  ; add header keywords
  if size(header,/type) ne 0 then sxaddpar,header,'CRREJ_JL','epoxi_crrej.pro','Program for cosmic ray removal (JYL)'

  ; restore overscan area
  ; We find the sizes of the edges based on imgmode.
  ; Notes: The number of restored raws and lines in the edges is
  ; greater than that in Table IV on page 71 in Space Science 
  ; Reviews, v. 117, Nos. 1-2, 2005 (e.g. 8 for mode=1),
  ; because e.g. for mode=1 9th row and 9th line are bright,
  ; 10th row and 10th line can have a lot of bright pixels,
  ; and 11th line can have some bright pixels which are not cosmic rays.
  ; The values of iy presented below were suggested 
  ; by Dennis Wellnitz. Dennis suggested to use ix=iy-1, 
  ; but for mode=3 I found that the 6th column also can contain a lot of 
  ; deleted pixels. So below the values of ix were taken to be
  ; equal to iy suggested by Dennis. Probably for some modes
  ; the values of ix can be the same as it was suggested to Dennis,
  ; but as there are some examples which give larger edges, I took ix=iy.
  ; I will be glad to get any comments about the sizes of the edges
  ; which must be restored in order to find true values.
  ; For what modes can we take ix=iy-1?
  ; Depending on one's problem, one can consider other values of ix and iy.
  if keep then begin
    ; gather header infomration
    imgmode=sxpar(header,'imgmode')
    inst=sxpar(header, 'INSTRUME')
    nx=sxpar(header,'naxis1')-1
    ny=sxpar(header,'naxis2')-1  ; size of the image [0:ny]

    ix0=0  ;;Number of rows to delete on the left
    ix1=0  ;;the right
    iy0=0  ;;the bottom
    iy1=0  ;;the top
    if strtrim(inst,2) EQ 'HRIIR' then begin     ;;IR instrument
      switch imgmode of
        1:
        5:begin
            ix0=3
            iy0=3
            iy1=1
            break
          end
        2:
        3:begin
            ix0=3
            iy0=2
            iy1=2
            break
          end
        4:
        6:begin
            ix0=5
            iy0=5
            iy1=1
            break
          end
      endswitch
      ix1=ix0
    endif else begin    ;;VIS instruments
      switch imgmode of
        1:
        9: BEGIN
             ix0=11 & break
           END
	2:
	3: BEGIN
		ix0=6 & break
	END
	5:
	6: BEGIN
		ix0=4 & break
	END
	7: BEGIN 
		ix0=2 & break
	END
	8: BEGIN
		ix0=1 & break
	END
	endswitch

	ix1=ix0
	iy0=ix0
	iy1=ix0
    endelse

    ; restore overscan
    tmp = input
    tmp[ix0:nx-ix1,iy0:ny-iy1] = out[ix0:nx-ix1,iy0:ny-iy1]
    out = tmp
    tmp = mask
    mask[*] = 2
    mask[ix0:nx-ix1,iy0:ny-iy1] = tmp[ix0:nx-ix1,iy0:ny-iy1]

  endif

  ; return output image
  return, out

end
