;  The size of the region
;    being fit by the parabola starts at csize pixels centered on the
;    initial point approximated by clicking the cursor on the image,
;    then contracts to 3 pixels as the center is found.  
;  Finds the x center first, then searches along that column to find
;    the y center, and iterates until there is no change.
;
;   Note: the LMfit routine frequently gives errors that suggest the
;         fit did not converge. For this reason, the fits are plotted
;         to give the user confidence that the centers are real.
;
;  Input parameters:
;     image      real  image for which center is to be found
;
;  Optional keywords:
;     xi, yi     real  initial guess for the center point (if they are
;                      not given, the image is displayed and the
;                      cursor is used to click on the approximate center
;     min,max    real  min and max values to be used in displaying the image
;     kern       int   number of pixels on each side of the center in
;                      fitting the parabola
;     nodisp     flag  flag telling routine to not re-display the
;                      image (for use in other routines that have
;                      already displayed it)
;     centroid   flag  Skip the parabolic fit and do a centroid fit
;     error      int   returned value to pass information
;                      -1 for solution not converging
;                       1 for no measurement (quit)
;                       100 for clicking 'c'
;                       101 for clicking 'x'
;                       200 for fit converging outside the image
;
;  Output:
;     x0, y0     real  center found by inverse parabolic interpolation
;                      or by position of the cursor (if user so desires)
;
;  Revisions:
;   10/20/10:  TLF - Made a lot of adjustments to try and get the MRI
;                    measurements to keep from drifting
;
;

pro center_par, imageorig, x0, y0, xi=xi, yi=yi, kern=kern, $
                min=min,max=max, noplot=noplot, nodisp=nodisp,error=error,$
                centroid=centroid

on_error, 2

; if not enough parameters, then print usage message
if n_params() lt 3  then begin
   print, 'Usage:  center_par, image, x0, y0, xi=xi, yi=yi, min=min,',$
           ' max=max, /nodisp'
   return
endif

coarse = 0
image = 1.0*imageorig
cr_err = 1
ky_err = ''
imsize=size(image)
subarr = 0
m1a = nint(0.25*imsize(1))
m1b = nint(0.75*imsize(1))
m2a = nint(0.25*imsize(2))
m2b = nint(0.75*imsize(2))

; if initial guess is not included, then display image & wait for
;   cursor click to get initial guess of the center
if N_elements(min) NE 1 then min=min(image)
if N_elements(max) NE 1 then max=max(median(image[m1a:m1b,m2a:m2b],5))

if not keyword_set(xi) or not keyword_set(yi) then begin
   coarse = 1
   if not keyword_set(nodisp) then dispim, image, lhist=1, uhist=99.5


; commands for platforms that accept mouse clicks for the cursor position
;   print,'----------------------------------------------------------'
;   print,'Click the middle button (or CTRL+Left button) to get an  '
;   print,'   expanded image for defining the centroid with the cursor'
;   print,'               OR '
;   print,'Click the left button on the center of the object'
;   print,'   for a parabolic solution'
;   print,'               OR '
;   print,'Click the right button to escape with no measurement'
;   print,'----------------------------------------------------------'

;  For the mac, need to use key presses
   print,'----------------------------------------------------------'
   print,'Press the "x" to get an expanded image for defining the '
   print,'   centroid with the cursor'
   print,'               OR '
   print,'Press "c" to center on the object under the cursor'
   print,'               OR '
   print,'Press "q" to escape with no measurement'
   print,'----------------------------------------------------------'
   goto, MEASURE

REMEASURE:
   wset,16
   wshow,16
   if subarr eq 1 then goto, LOOP2
   print,'****************'
   print,' The position is more than 10 pixels from the initial position' 
   print,'   or the solution is not converging'
   print,' Redo or click the middle button to use a given pixel as the center'
   

MEASURE:
   wset,16
   wshow,16
   LOOP: $
;sv_err = !ERR
     ky_err = ''
;     !ERR = 0
     cursor,xi,yi,2,/CHANGE,/DEVICE
;     cursor,xi,yi,2,/down,/DEVICE
     ky_err = get_kbrd(0)  
     case ky_err of
       'x':  goto,ENDLOOP
       'c':  goto,ENDLOOP
       'q':  goto,BADLOOP
       else: goto,LOOP
     endcase
;     cr_err = !ERR
;     if cr_err eq 1 or cr_err eq 2 then goto,ENDLOOP
;     if cr_err eq 4 then goto,BADLOOP
ENDLOOP:
endif

x1 = xi
y1 = yi

;if the middle button was pushed, use that value, redisplay a subarray
;around the desired region, and wait for another button click, which
;will be used as the center.  If the left button is clicked, use the center
;and go to the accurate parabola fitting routine.  If the middle button is
;clicked, use that point as the center

if cr_err eq 2 or ky_err eq 'x' then begin
   wset,16
   wshow,16
   coarse = 0
   max2 = max(image[xi-5:xi+5,yi-5:yi+5],min=min2)
   dispim, image[xi-25:xi+25,yi-25:yi+25],lhist=1,uhist=99.5,exp=10
   image_section,image,xi,yi,subimage,/nocenter,frame=[51,51]
   dispim, subimage,lhist=1,uhist=99.5,exp=10
   print,''   
   print,' Zooming in on the area centered on pixel ',xi,yi
   print,''
;   print,'----------------------------------------------------------'
;   print,'Click the middle button (or CTRL+Left button) to define '
;   print,'   an integer pixel to use as the center'
;   print,'               OR '
;   print,'Click the left button to define the starting point to use'
;   print,'   for a parabolic fitting solution'
;   print,'               OR '
;   print,'Click the right button to escape with no measurement'
;   print,'----------------------------------------------------------'
 
   print,'----------------------------------------------------------'
   print,'Press "x" to define an integer pixel to use as the center'
   print,'               OR '
   print,'Press "c" to define the starting point for centering'
   print,'               OR '
   print,'Press "q" to escape with no measurement'
   print,'----------------------------------------------------------'

   LOOP2: $ sv_err = !ERR
     !ERR = 0
     ky_err = ''
     cursor,xb,yb,2,/CHANGE,/DEVICE
     ky_err = get_kbrd(0)                               
     cr_err = !ERR
     case ky_err of
       'x':  goto,ENDLOOP2
       'c':  goto,ENDLOOP2
       'q':  goto,BADLOOP
       else: goto,LOOP2
     endcase
;     if cr_err eq 1 or cr_err eq 2 then goto,ENDLOOP2
;     if cr_err eq 4 then goto,BADLOOP
   ENDLOOP2:

   if strcmp(ky_err,'c') then error=100
   if strcmp(ky_err,'x') then error=101
;    find the center relative to the subimage array -- (xb-5)/10 deals
;    with the problem of expanding the image for display
   x0=xi-25+nint((xb-5)/10.)
   y0=yi-25+nint((yb-5)/10.)
   subarr = 1
   if cr_err eq 1 or ky_err eq 'c' then $
     if keyword_set(centroid) then goto,CENTROID else goto,ACCFIT $
     else goto, PRINTOUT
endif

; If 'c' ws clicked on full-sized image, come here
print,'Clicked at ',xi,yi
if keyword_set(centroid) then goto,CENTROID  

; set up testing values
if coarse eq 0 then holdsize= 3 else holdsize=10

RESTART:
csize=holdsize ;set to 3

; find the location of the brighest pixel in the subframe around the cursor,
; to use as the starting position
xa = round(x1)-csize
xc = round(x1)+csize
ya = round(y1)-csize
yc = round(y1)+csize
maxpix = max(image[xa:xc,ya:yc])
pix = where(image[xa:xc,ya:yc] eq maxpix)
x0 = pix mod (2*csize+1)
y0 = (pix-x0)/(2*csize+1)
x0 = xa+x0
y0 = ya+y0

;  come here to get a more accurate fit using a parabola segment
ACCFIT:
if keyword_set(kern) then kernal=kern
if not keyword_set(kern) then kernal=3 ;kernal 
if kernal eq 1 then goto, PRINTOUT
kernal = nint(kernal)
xvec = intarr(2*kernal+1)
yvec = intarr(2*kernal+1)

; fit the x direction
xa = x0-kernal
xc = x0+kernal
for i1 = 0,2*kernal do  xvec[i1] = nint(xa+i1)
secx = 1.0*image[nint(xa):nint(xc),nint(y0)]

p2 = -0.25*(xa-x0)^2/(image(nint(xa),nint(y0))-image(nint(x0),nint(y0)))
param = [x0, p2, image[nint(x0),nint(y0)]]
par = parabola(xvec,param)

; call the L-M minimization routine to fit for the parabola
newpar = lmfit(xvec,secx,param,function_name='parabola',/double)
x0 = param[0]
if x0 gt imsize(1) or x0 lt 0 then begin
   print,'X direction did not converge in parabolic fit'
   goto,CENTROID
endif
x200 = (findgen(200)-100)*kernal/100
parx = parabola(x200+x0,param)

;---------------------------------------
; fit the y direction
ya = y0-kernal
yc = y0+kernal
for i1 = 0,2*kernal do  yvec[i1] = nint(ya+i1)
secy = image[nint(x0),nint(ya):nint(yc)]
p2 = -0.25*(ya-y0)^2/(image(nint(x0),nint(ya))-image(nint(x0),nint(y0)))
param = [y0, p2, image[nint(x0),nint(y0)]]

; call the L-M minimization routine to fit for the parabola
newpar = lmfit(yvec,secy,param,function_name='parabola',/double)
y0 = param[0]
if y0 gt imsize(2) or y0 lt 0 then begin
   print,'Y direction did not converge in parabolic fit'
   goto,CENTROID
endif
pary = parabola(x200+y0,param)
if parx[0] gt parx[100] or pary[0] gt pary[100] then goto, REMEASURE

; plot the parabola to show the fit
if keyword_set(noplot) then goto,PRINTOUT
window,13,xsize=400,ysize=300
xpos=[0.08,0.1,0.48,0.9]
maxv=max(secx,min=minv)
plot,xvec,secx,position=xpos,title='Fit in X direction',$
   yrange=[0.99*min(secx),1.02*maxv],ystyle=1
oplot,x200+x0,parx,thick=2
xpos=[0.55,0.1,0.95,0.9]
plot, yvec,secy,position=xpos,title='Fit in Y direction',$
    /noerase,yrange=[0.99*min(secy),1.02*maxv],ystyle=1
oplot, x200+y0,pary,thick=2

xdif = x0-xi
ydif = y0-yi
diff = sqrt(xdif^2+ydif^2)
if (diff gt 2*kernal) then goto,REMEASURE else goto,PRINTOUT

CENTROID:
print,''
print, '   Moving to centroid fitting section '
if keyword_set(kern) then kernal=kern
if not keyword_set(kern) then kernal=2
cntrd,image,xi,yi,x0,y0,2*kernal,/silent
xi=x0
yi=y0
cntrd,image,xi,yi,x0,y0,2*kernal,/silent
if x0 eq -1 then begin
  print,'Centroid solution did not converge - retrying'
  if try1 eq 0 then begin
     try1 = 1
     goto, REMEASURE
  endif
  error = -1
  return
endif

PRINTOUT:
print,' '
if ky_err ne 'x' then print,' Successfully fit the center:'
if ky_err ne 'x' then print,format='("Initial position:   ",i4,7x,i4)',xi,yi
print,format='("Center is at:       ",f8.3,3x,f8.3)',x0,y0
print,' '
return

BADLOOP:
print,'Escaping with no measurement'
error=1
return
end

